/*****************************************************************************\
 *  scrub.c,v 1.10 2003/04/11 16:13:39 garlick Exp
 *****************************************************************************
 *  Copyright (C) 2001-2002 The Regents of the University of California.
 *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
 *  Written by Jim Garlick <garlick@llnl.gov>.
 *  UCRL-CODE-2003-006.
 *  
 *  This file is part of Scrub, a program for erasing disks.
 *  For details, see <http://www.llnl.gov/linux/scrub/>.
 *  
 *  Scrub is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *  
 *  Scrub is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with Scrub; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
\*****************************************************************************/

/* Scrub a raw disk or plain file.
 */

#if defined(linux) || defined(sun)
#define _LARGEFILE_SOURCE 
#define _FILE_OFFSET_BITS 64
#endif

#if defined(_AIX)
#define _LARGE_FILES
#include <sys/mode.h>
#endif

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <libgen.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "genrand.h"

#define RANDOM	'r'

static unsigned char pattern[] = { 0, 0xff, 0xaa, RANDOM, 0x55 };

typedef enum { false, true } bool;

static long long scrubfile(int, unsigned char *, size_t, unsigned char, 
        size_t);
static long long verifyfile(int, unsigned char *, size_t, unsigned char, 
        size_t);
static char *pat2str(unsigned char);
static void activity(char *msg, long long megabytes);
static void usage(void);
static char *xbasename(char *);
static void writedentry(char *, unsigned char);

static bool quiet= false;     /* quiet mode */
static bool dopt = false;     /* debugging, no action */
static bool ropt = false;     /* include random pass */
static bool xopt = false;     /* write an extra block past size of file */
static bool Xopt = false;     /* create a new file and keep appending forever */
static bool Dopt = false;     /* scrub directory entry too */
static bool special = false;  /* target is a special file */
static char *dev = NULL;      /* device/file to scrub */

/* 
 * If scrubbing a special file, the number of filesystem blocks to write
 * in one chunk.
 */
#define NUMFSBLOCKS 1024

static char *progname;

int main(int argc, char *argv[])
{
    struct stat sb;
    unsigned char *buf;
    int c, fd, pass, numpasses;
    size_t chunksize, numchunks; 
    long long sz, last_sz = -1;
    extern int optind;
    extern char *optarg;

    /* 
     * Handle arguments.
     */
    progname = xbasename(argv[0]);
    while ((c = getopt(argc, argv, "DXxdrf:")) != EOF) {
        switch (c) {
        case 'd':
            dopt = true;
            break;
        case 'r':
            ropt = true;
            break;
        case 'x':
            xopt = true;
            break;
        case 'X':
            Xopt = true;
            break;
        case 'D':
            Dopt = true;
            break;
        default:
            usage();
        }
    }
    if (argc - optind != 1)
        usage();
    dev = argv[optind++];
    if (Xopt && xopt)
        usage();

    quiet = !isatty(1);

    /*
     * Open and stat the device/file.
     * (if nonexistant and -X, create)
     */
    if (dev == NULL)
        usage();
    fd = open(dev, O_RDWR | (Xopt ? O_CREAT : 0), 0644);
    if (fd == -1) {
        perror(dev);
        exit(1);
    }
    if (fstat(fd, &sb) == -1) {
        fprintf(stderr, "%s: can't stat %s\n", progname, dev);
        exit(1);
    }

    /*
     * Determine 'chunksize' (size of read/write's to be attempted)
     * and 'numchunks', the number of chunks to read/write.  If character
     * special file, numchunks is not used;  we will just go until we get
     * an error, which is assumed to be an attempt to read/write past the 
     * end of the device.
     */
#ifdef linux
    /* linux doesn't have character files for disk devices */
    if (S_ISBLK(sb.st_mode)) {  /* special file */
#else
    if (S_ISCHR(sb.st_mode)) {  /* special file */
#endif
        if (xopt) {
            fprintf(stderr, "%s: -x only works on plain files\n",
                    progname);
            exit(1);
        }
        chunksize = sb.st_blksize * NUMFSBLOCKS;
        numchunks = 0;
        printf("%s: will write %d-byte chunks on %s\n",
               progname, chunksize, dev);
        special = true;
    } else if (S_ISREG(sb.st_mode)) {   /* just a file */
        chunksize = sb.st_blksize;
        numchunks = sb.st_size / chunksize;     /* approximate length */
        if (xopt)
            numchunks++;
        if (Xopt)
            printf("%s: will write %d-byte chunks on %s\n", 
                    progname, chunksize, dev);
        else
            printf("%s: will write %d %d-byte chunks on %s\n", 
                    progname, numchunks, chunksize, dev);
    } else {
        fprintf(stderr, "%s: %s is not a plain file or char device\n",
                progname, dev);
        exit(1);
    }

    /* 
     * Allocate one buffer which we reuse in successive calls to scrubfile.
     * If writing a random pass, seed the RNG.
     */
    buf = (unsigned char *) malloc(chunksize);
    if (buf == NULL) {
        perror("malloc");
        exit(1);
    }
    if (ropt)
        initrand();

    /* 
     * Do the write passes.
     */
    numpasses = sizeof(pattern) / sizeof(unsigned char);
    for (pass = 0; pass < numpasses; pass++) {
        if (pattern[pass] == RANDOM && !ropt)
            continue;
        if (dopt) {
            printf("%s: pattern %s\n", progname, pat2str(pattern[pass]));
            continue;
        }
        sz = scrubfile(fd, buf, chunksize, pattern[pass], numchunks);
        if (last_sz != -1 && sz != last_sz) {
            fprintf(stderr,
                    "%s: this pass=%lld bytes, last pass=%lld bytes\n",
                    progname, sz, last_sz);
            exit(1);
        }
        last_sz = sz;
    }

    /* 
     * Verify that the last pattern written is what we see on the disk.
     */
    if (!dopt) {
        printf("%s: verify pattern 0x%x\n", progname, pattern[pass - 1]);
        sz = verifyfile(fd, buf, chunksize, pattern[pass - 1], numchunks);
        if (sz != last_sz) {
            fprintf(stderr,
                    "%s: verify pass=%lld bytes, last pass=%lld bytes\n",
                    progname, sz, last_sz);
            exit(1);
        }
    }

    /*
     * Scrub directory entry too?
     */
    if (Dopt) {
        for (pass = 0; pass < numpasses; pass++) {
            if (pattern[pass] == RANDOM && !ropt)
                continue;
            if (pattern[pass] == 0)
                continue;
            if (dopt) {
                printf("%s: pattern %s on dentry\n", 
                        progname, pat2str(pattern[pass]));
                continue;
            }
            writedentry(dev, pattern[pass]);
        }
    }
    exit(0);
}


static long long scrubfile(int fd, unsigned char *buf, size_t chunksize, 
        unsigned char val, size_t numchunks)
{
    int rv = 0, count = 0;
    bool randomize = (val == RANDOM);
    long long bytes_written = 0;

    /* apply pattern to buffer */
    if (randomize)
        genrand(buf, chunksize);
    else
        memset(buf, val, chunksize);

    /* seek to beginning */
    if (lseek(fd, 0, SEEK_SET) == -1) {
        perror("lseek");
        fflush(stderr);
        exit(1);
    }

    /* write until we hit the end of the device or file */
    while (Xopt || special || count < numchunks) {
        rv = write(fd, buf, chunksize);
        if (rv < 0)
            perror("write");
        if (rv > 0) {
            bytes_written += rv;
            activity(pat2str(val), bytes_written/(1024*1024));
        }
        if (rv < 0 || rv < chunksize)
            break;
        count++;
        if (randomize) {
            churnrand();
            genrand(buf, chunksize);
        }
    }

    /* report */
    printf("%s(%s): wrote %d chunks on %s\n",
           progname, pat2str(val), count, dev);
    if (rv > 0 && rv != chunksize) {
        printf("%s(%s): plus one partial chunk (%d bytes) on %s\n",
               progname, pat2str(val), rv, dev);
    }

    if (!special) {
        printf("scrub: flushing file to disk with fsync\n");
        fflush(stdout);
        fsync(fd);
    }

    return bytes_written;
}

static long long verifyfile(int fd, unsigned char *buf, size_t chunksize, 
        unsigned char val, size_t numchunks)
{
    int rv = 0, count = 0;
    unsigned char *buf2;
    long long bytes_read = 0;

    if ((buf2 = (unsigned char *) malloc(chunksize)) == NULL) {
        perror("malloc");
        exit(1);
    }

    /* seek to beginning */
    if (lseek(fd, 0, SEEK_SET) == -1) {
        perror("lseek");
        exit(1);
    }

    /* apply pattern to buffer */
    memset(buf, val, chunksize);

    /* read until we hit end of device or file */
    while (Xopt || special || count < numchunks) {
        rv = read(fd, buf2, chunksize);
        if (rv > 0 && memcmp(buf, buf2, rv) != 0) {
            fprintf(stderr, "%s(%s): verification failed at %schunk %d on %s\n",
                    progname, pat2str(val), 
                    rv == chunksize ? "" : "partial ", count, dev);
            exit(1);
        }
        if (rv > 0) {
            bytes_read += rv;
            activity(pat2str(val), bytes_read/(1024*1024));
        }
        if (rv < chunksize)
            break;
        count++;
    }

    /* report */
    printf("%s(%s): verified %d chunks on %s\n",
           progname, pat2str(val), count, dev);
    if (rv > 0 && rv != chunksize) {
        printf("%s: plus one partial chunk (%d bytes) on %s\n",
               progname, rv, dev);
    }

    return bytes_read;
}

static void usage(void)
{
    int pass;

    fprintf(stderr, "Usage: %s [-d] [-r] [-x|-X] file\n", progname);
    fprintf(stderr, "\t-r includes random pass (slow!)\n");
    fprintf(stderr,
            "\t-d inhibits writing (just tell what we're going to do)\n");
    fprintf(stderr, "\t-x write size of file + one filesystem block\n");
    fprintf(stderr,
            "\t-X create file and keep writing until write fails\n");
    fprintf(stderr,
            "\t-D scrub directory entry too using same algorithm\n");
    fprintf(stderr,
            "\tfile is raw special file or existing regular file\n");

    fprintf(stderr, "\tPasses are: ");
    for (pass = 0; pass < sizeof(pattern) / sizeof(unsigned char); pass++)
        printf("%d:[%s] ", pass, pat2str(pattern[pass]));
    printf("%d:[verify]\n", pass);

    exit(1);
}

static char *pat2str(unsigned char pat)
{
    static char str[255];

    if (pat == RANDOM)
        sprintf(str, "random");
    else
        sprintf(str, "0x%x", pat);
    return str;
}

static void activity(char *msg, long long megabytes)
{
    char prop[4] = { '|', '/', '-', '\\' };
    static int prop_count = 0;
    long long last_megabytes = -1;

    if (!quiet && last_megabytes != megabytes) {
        if (megabytes < 1000) {
            printf("%s(%s): %c %-3.3dMB\r", progname, msg, 
                    prop[prop_count++ % 4], (int)megabytes);
        } else if (megabytes < 1000*1000) {
            printf("%s(%s): %c %-3.3dGB\r", progname, msg, 
                    prop[prop_count++ % 4], (int)(megabytes/1024));
        } else {
            printf("%s(%s): %c %-3.3dTB\r", progname, msg, 
                    prop[prop_count++ % 4], (int)(megabytes/(1024*1024)));
        }
        last_megabytes = megabytes;
        fflush(stdout);
    }
}

static char *xbasename(char *path)
{
    char *p = strrchr(path, '/');

    return (p ? p + 1 : path);
}

static void writedentry(char *name, unsigned char pat)
{
    char *old = strdup(name);
    char *base = xbasename(name);

    memset(base, pat, strlen(base));
    if (rename(old, name) == -1) {
        perror("rename");
        exit(1);
    }
    free(old);
    printf("%s(%s): wrote dentry...sync\n", progname, pat2str(pat));
    sync();
    sleep(1);
}


/*
 * vi:tabstop=4 shiftwidth=4 expandtab
 */
